import uuid
from unittest.mock import MagicMock, patch

from django.urls import reverse
from django.utils import timezone as DjangoTimeZone
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessMode
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import fake_now
from arkindex.training.models import Dataset, DatasetState
from arkindex.users.models import Role, User

# Using the fake DB fixtures creation date when needed
FAKE_CREATED = "2020-02-02T01:23:45.678000Z"


class TestDatasetsAPI(FixtureAPITestCase):
    @classmethod
    @patch.object(DjangoTimeZone, "now", fake_now)
    def setUpTestData(cls):
        super().setUpTestData()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        cls.task = cls.process.tasks.first()
        cls.dataset_creator = User.objects.create(email="adam@nerv.co.jp", display_name="Adam", verified_email=True)
        cls.private_corpus = Corpus.objects.create(name="private", public=False)
        cls.read_user = User.objects.get(email="user3@user.fr")
        cls.write_user = User.objects.get(email="user2@user.fr")
        cls.dataset = Dataset.objects.get(name="First Dataset")
        cls.dataset2 = Dataset.objects.get(name="Second Dataset")
        cls.process.datasets.set((cls.dataset, cls.dataset2))
        cls.private_dataset = Dataset.objects.create(name="Private Dataset", description="Dead Sea Scrolls", corpus=cls.private_corpus, creator=cls.dataset_creator)
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.page1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.page2 = cls.corpus.elements.get(name="Volume 1, page 1v")
        cls.page3 = cls.corpus.elements.get(name="Volume 1, page 2r")

    # ListCorpusDatasets

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_list_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_list_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_list_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.private_corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have guest access to this corpus."})

    def test_list(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["results"], [
            {
                "name": "First Dataset",
                "description": "dataset number one",
                "id": str(self.dataset.pk),
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "state": "open",
                "creator": "Test user",
                "task_id": None,
                "corpus_id": str(self.corpus.id),
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED,
            },
            {
                "name": "Second Dataset",
                "description": "dataset number two",
                "id": str(self.dataset2.pk),
                "sets": ["training", "test", "validation"],
                "set_elements": None,
                "state": "open",
                "creator": "Test user",
                "task_id": None,
                "corpus_id": str(self.corpus.id),
                "created": FAKE_CREATED,
                "updated": FAKE_CREATED,
            }
        ])

    # CreateDataset

    def test_create_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_create_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_create_corpus_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_create_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.private_corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to this corpus."})

    def test_create_name_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field is required."]})

    def test_create_name_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_name_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "        ", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."]})

    def test_create_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                    "description": "My dataset for my experiments."
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_create_name_already_exists_in_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "First Dataset",
                    "description": "My dataset for my experiments."
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_create_description_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field is required."]})

    def test_create_description_empty(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": ""},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field may not be blank."]})

    def test_create_description_blank(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "             "},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field may not be blank."]})

    def test_create(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments."},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ["training", "test", "validation"],
            "set_elements": {
                "training": 0,
                "test": 0,
                "validation": 0,
            },
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_state_ignored(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "state": "complete"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertEqual(created_dataset.state, DatasetState.Open)
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ["training", "test", "validation"],
            "set_elements": {"test": 0, "training": 0, "validation": 0},
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_sets(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "sets": ["a", "b", "c", "d"]},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        created_dataset = Dataset.objects.get(name="My dataset")
        self.assertDictEqual(response.json(), {
            "id": str(created_dataset.id),
            "name": "My dataset",
            "description": "My dataset for my experiments.",
            "sets": ["a", "b", "c", "d"],
            "set_elements": {"a": 0, "b": 0, "c": 0, "d": 0},
            "state": "open",
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "created": created_dataset.created.isoformat().replace("+00:00", "Z"),
            "updated": created_dataset.updated.isoformat().replace("+00:00", "Z"),
        })

    def test_create_sets_length(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "sets": []},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Either do not specify set names to use the default values, or specify a non-empty list of names."]})

    def test_create_sets_unique_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "sets": ["a", "a", "b"]},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Set names must be unique."]})

    def test_create_sets_blank_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={"name": "My dataset", "description": "My dataset for my experiments.", "sets": ["     ", " ", "b"]},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets":
                {
                    "0": ["This field may not be blank."],
                    "1": ["This field may not be blank."]
                }
        })

    def test_create_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:corpus-datasets", kwargs={"pk": self.corpus.pk}),
                data={
                    "name": "My Dataset",
                    "description": "My dataset for my experiments.",
                    "sets": ["unit-00", "Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor."]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets": {
                "1": ["Ensure this field has no more than 50 characters."]
            }
        })

    # UpdateDataset

    def test_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["unit-01", "unit-00", "unit-02"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_update_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["unit-01", "unit-00", "unit-02"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_update_requires_write_corpus(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["unit-01", "unit-00", "unit-02"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

    def test_update_doesnt_exist(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["unit-01", "unit-00", "unit-02"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                    "description": "My dataset for my experiments.",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Another Dataset",
                    "description": "My dataset for my experiments.",
                    "sets": self.dataset.sets,
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_update_requires_all_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"description": ["This field is required."]})

    def test_update_add_sets(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.dataset.task_id)
        with self.assertNumQueries(11):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": self.dataset.sets + ["unit-01", "unit-02"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, "Shin Seiki Evangelion")
        self.assertEqual(self.dataset.description, "Omedeto!")
        self.assertListEqual(self.dataset.sets, ["training", "test", "validation", "unit-01", "unit-02"])
        self.assertIsNone(self.dataset.task_id)

    def test_update_remove_sets(self):
        """
        It is possible to remove many sets, no elements are moved
        """
        self.client.force_login(self.user)
        dataset_elt = self.dataset.dataset_elements.create(element=self.page1, set="training")
        with self.assertNumQueries(11):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    # train and validation sets are removed
                    "sets": ["test"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, "Shin Seiki Evangelion")
        self.assertEqual(self.dataset.description, "Omedeto!")
        self.assertListEqual(self.dataset.sets, ["test"])
        self.assertIsNone(self.dataset.task_id)
        dataset_elt.refresh_from_db()
        self.assertEqual(dataset_elt.element, self.page1)
        self.assertEqual(dataset_elt.dataset, self.dataset)
        # Element is still linked to a set that does not exist anymore on the dataset
        self.assertEqual(dataset_elt.set, "training")

    def test_update_sets_update_single_set(self):
        """
        It is possible to rename a single set
        """
        self.client.force_login(self.user)
        self.dataset.dataset_elements.create(element_id=self.page1.id, set="training")
        self.dataset.dataset_elements.create(element_id=self.page2.id, set="validation")
        self.dataset.dataset_elements.create(element_id=self.page3.id, set="validation")
        with self.assertNumQueries(12):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    # validation set is renamed to AAAAAAA
                    "sets": ["test", "training", "AAAAAAA"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, "Shin Seiki Evangelion")
        self.assertEqual(self.dataset.description, "Omedeto!")
        self.assertListEqual(self.dataset.sets, ["test", "training", "AAAAAAA"])
        self.assertIsNone(self.dataset.task_id)
        self.assertQuerysetEqual(
            self.dataset.dataset_elements.values_list("set", "element__name").order_by("element__name"),
            [
                ("training", "Volume 1, page 1r"),
                ("AAAAAAA", "Volume 1, page 1v"),
                ("AAAAAAA", "Volume 1, page 2r"),
            ]
        )

    def test_update_sets_ambiguous(self):
        """
        No more than one set can be updated
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["test", "AAAAAAA", "BBBBBBB"],
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"sets": ["Updating those sets is ambiguous because several have changed."]})

    def test_update_sets_length(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": []
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Either do not specify set names to use the default values, or specify a non-empty list of names."]})

    def test_update_sets_unique_names(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "sets": ["unit-01", "unit-01", "unit-02"]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"sets": ["Set names must be unique."]})

    def test_update_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "My Dataset",
                    "description": "My dataset for my experiments.",
                    "sets": ["unit-00", "Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor."]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets": {
                "1": ["Ensure this field has no more than 50 characters."]
            }
        })

    def test_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "",
                    "description": "          ",
                    "sets": "",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["This field may not be blank."],
            "description": ["This field may not be blank."],
            "sets": ['Expected a list of items but got type "str".'],
        })

    def test_update_all_errors(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "",
                    "description": "          ",
                    "sets": ["unit-00", "unit-00", "unit-02"]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "name": ["This field may not be blank."],
            "description": ["This field may not be blank."],
            "sets": ["Set names must be unique."]
        })

    def test_update_state_requires_ponos_auth(self):
        self.client.force_login(self.user)
        self.dataset.state = DatasetState.Building
        self.dataset.save()
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "complete"
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "state": ["Ponos task authentication is required to update the state of a Dataset."]
        })

    def test_update_ponos_task_state_update(self):
        self.dataset.state = DatasetState.Building
        self.dataset.save()
        with self.assertNumQueries(11):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "complete"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(response.json()["set_elements"], {"test": 0, "training": 0, "validation": 0})
        self.assertEqual(self.dataset.state, DatasetState.Complete)
        self.assertEqual(self.dataset.task_id, self.task.id)

    def test_update_ponos_task_state_forbidden(self):
        """Dataset's state update is limited to specific transitions"""
        op, build, complete, error = [DatasetState[state] for state in ("Open", "Building", "Complete", "Error")]
        states = {
            (op, op): True,
            (op, build) : True,
            (op, complete) : False,
            (op, error): False,
            (build, op): False,
            (build, build) : True,
            (build, complete) : True,
            (build, error): True,
            (error, op): False,
            (error, build) : True,
            (error, complete) : False,
            (error, error): True
        }
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.put(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                    data={
                        "name": self.dataset.name,
                        "description": self.dataset.description,
                        "state": state_to.value,
                    },
                    format="json",
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_update_complete_forbidden(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        self.client.force_login(self.user)
        for new_state in DatasetState:
            with self.subTest(new_state=new_state):
                with self.assertNumQueries(6):
                    response = self.client.put(
                        reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                        data={
                            "name": "AA",
                            "description": "BB",
                            "sets": self.dataset.sets + ["CC"],
                            "state": new_state.value,
                        },
                        format="json"
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertListEqual(
                    response.json(),
                    ["This dataset is in complete state and cannot be modified anymore."],
                )

    def test_update_ponos_task_state_requires_dataset_in_process(self):
        self.process.process_datasets.all().delete()
        with self.assertNumQueries(7):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "building"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": ["A task can only update the state of one of the datasets of its process."]
        })

    def test_update_ponos_task_bad_state(self):
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "name": "Shin Seiki Evangelion",
                    "description": "Omedeto!",
                    "state": "decommissioned"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ["Value is not of type DatasetState"]})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    # PartialUpdateDataset

    def test_partial_update_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_partial_update_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_partial_update_requires_write_corpus(self):
        self.client.force_login(self.read_user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={"name": "Shin Seiki Evangelion"},
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have contributor access to corpus Unit Tests."})

    def test_partial_update_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "In elementum leo et lectus finibus, eget lobortis ligula posuere. Vivamus finibus quis sapien in mattis. Cras in libero.",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["Ensure this field has no more than 100 characters."]})

    def test_partial_update_name_already_exists_in_corpus(self):
        Dataset.objects.create(name="Another Dataset", description="A set of data", corpus=self.corpus, creator=self.dataset_creator)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "Another Dataset",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["A dataset with this name already exists in this corpus."]})

    def test_partial_update(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "description": "Omedeto!",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)
        self.assertEqual(self.dataset.name, "First Dataset")
        self.assertEqual(self.dataset.description, "Omedeto!")
        self.assertListEqual(self.dataset.sets, ["training", "test", "validation"])

    def test_partial_update_empty_or_blank_description_or_name(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "name": "                  ",
                    "description": "",
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"name": ["This field may not be blank."], "description": ["This field may not be blank."]})

    def test_partial_update_requires_ponos_auth(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "state": "building"
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {
            "state": ["Ponos task authentication is required to update the state of a Dataset."]
        })

    def test_partial_update_ponos_task_state_update(self):
        with self.assertNumQueries(11):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "state": "building"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Building)

    def test_partial_update_ponos_task_state_requires_dataset_in_process(self):
        self.process.process_datasets.all().delete()
        with self.assertNumQueries(7):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "state": "building"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "state": ["A task can only update the state of one of the datasets of its process."]
        })

    def test_partial_update_ponos_task_bad_state(self):
        with self.assertNumQueries(5):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                data={
                    "state": "decommissioned"
                },
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"state": ["Value is not of type DatasetState"]})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.state, DatasetState.Open)

    def test_partial_update_sets_name_too_long(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.patch(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                data={
                    "sets": ["unit-00", "Etiam accumsan ullamcorper mauris eget mattis. Ut porttitor."]
                },
                format="json"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "sets": {
                "1": ["Ensure this field has no more than 50 characters."]
            }
        })

    def test_partial_update_ponos_task_state_forbidden(self):
        """Dataset's state update is limited to specific transitions"""
        op, build, complete, error = [DatasetState[state] for state in ("Open", "Building", "Complete", "Error")]
        states = {
            (op, op): True,
            (op, build) : True,
            (op, complete) : False,
            (op, error): False,
            (build, op): False,
            (build, build) : True,
            (build, complete) : True,
            (build, error): True,
            (error, op): False,
            (error, build) : True,
            (error, complete) : False,
            (error, error): True
        }
        for ((state_from, state_to), expected) in states.items():
            with self.subTest(state_from=state_from, state_to=state_to):
                self.dataset.state = state_from
                self.dataset.save()
                response = self.client.patch(
                    reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    HTTP_AUTHORIZATION=f"Ponos {self.task.token}",
                    data={
                        "name": self.dataset.name,
                        "description": self.dataset.description,
                        "state": state_to.value,
                    },
                    format="json",
                )
                if expected:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    self.dataset.refresh_from_db()
                    self.assertEqual(self.dataset.state, state_to)
                else:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                    self.assertDictEqual(
                        response.json(),
                        {"state": [f"Transition from {state_from} to {state_to} is not allowed."]}
                    )

    def test_partial_update_complete_forbidden(self):
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        self.client.force_login(self.user)
        for new_state in DatasetState:
            with self.subTest(new_state=new_state):
                with self.assertNumQueries(6):
                    response = self.client.patch(
                        reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                        data={
                            "name": "AA",
                            "description": "BB",
                            "sets": self.dataset.sets + ["CC"],
                            "state": new_state.value,
                        },
                        format="json"
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertListEqual(
                    response.json(),
                    ["This dataset is in complete state and cannot be modified anymore."],
                )

    # RetrieveDataset

    def test_retrieve_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_retrieve_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_retrieve_requires_read_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.private_dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_retrieve_doesnt_exists(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_retrieve(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "id": str(self.dataset.id),
            "name": "First Dataset",
            "description": "dataset number one",
            "state": "open",
            "sets": ["training", "test", "validation"],
            "set_elements": {"test": 0, "training": 0, "validation": 0},
            "creator": "Test user",
            "task_id": None,
            "corpus_id": str(self.corpus.id),
            "created": FAKE_CREATED,
            "updated": FAKE_CREATED
        })

    def test_retrieve_task_id(self):
        self.client.force_login(self.user)
        self.dataset.task = self.task
        self.dataset.save()
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk})
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["task_id"], str(self.task.id))

    # DestroyDataset

    def test_delete_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_delete_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_delete_requires_write_corpus(self):
        self.client.force_login(self.write_user)
        with self.assertNumQueries(6):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have admin access to corpus Unit Tests."})

    def test_delete_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}),
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_delete(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Dataset.DoesNotExist):
            self.dataset.refresh_from_db()

    def test_delete_not_open(self):
        self.client.force_login(self.user)
        cases = [DatasetState.Building, DatasetState.Complete, DatasetState.Error]
        for state in cases:
            with self.subTest(state=state):
                self.dataset.state = state
                self.dataset.save()
                with self.assertNumQueries(5):
                    response = self.client.delete(
                        reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
                    )
                    self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
                self.assertDictEqual(response.json(), {"detail": "Only datasets in an open state can be deleted."})

    def test_delete_elements(self):
        """
        DestroyDataset also deletes DatasetElements
        """
        self.dataset.dataset_elements.create(element_id=self.vol.id, set="test")
        self.dataset.dataset_elements.create(element_id=self.vol.id, set="training")
        self.dataset.dataset_elements.create(element_id=self.page1.id, set="training")
        self.dataset.dataset_elements.create(element_id=self.page2.id, set="validation")
        self.dataset.dataset_elements.create(element_id=self.page3.id, set="validation")
        self.client.force_login(self.user)

        with self.assertNumQueries(9):
            response = self.client.delete(
                reverse("api:dataset-update", kwargs={"pk": self.dataset.pk}),
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        with self.assertRaises(Dataset.DoesNotExist):
            self.dataset.refresh_from_db()

        self.assertFalse(self.dataset.dataset_elements.exists())

        # No elements should have been deleted
        self.vol.refresh_from_db()
        self.page1.refresh_from_db()
        self.page2.refresh_from_db()
        self.page3.refresh_from_db()

    def test_list_elements_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(self.dataset.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list_elements_invalid_dataset_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(uuid.uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_elements_readable_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:dataset-elements", kwargs={"pk": str(self.private_dataset.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_list_elements_set_filter_wrong_set(self):
        self.dataset.dataset_elements.create(element_id=self.page1.id, set="test")
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse("api:dataset-elements", kwargs={"pk": str(self.dataset.id)}),
                data={"set": "aaaaa"}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": None,
            "next": None,
            "previous": None,
            "results": [],
        })

    def test_list_elements_set_filter(self):
        self.dataset.dataset_elements.create(element_id=self.page1.id, set="test")
        self.dataset.dataset_elements.create(element_id=self.page2.id, set="training")
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.pk}),
                data={"set": "training", "with_count": "true"},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        results = data.pop("results")
        self.assertDictEqual(data, {"count": 1, "next": None, "previous": None})
        self.assertEqual(len(results), 1)
        dataset_element = results[0]
        self.assertEqual(dataset_element["element"]["id"], str(self.page2.id))
        self.assertEqual(dataset_element["set"], "training")

    @patch("arkindex.documents.models.Element.thumbnail", MagicMock(s3_url="s3_url"))
    def test_list_elements(self):
        self.dataset.dataset_elements.create(element_id=self.vol.id, set="test")
        self.dataset.dataset_elements.create(element_id=self.page1.id, set="training")
        self.dataset.dataset_elements.create(element_id=self.page2.id, set="validation")
        self.dataset.dataset_elements.create(element_id=self.page3.id, set="validation")
        self.page3.confidence = 0.42
        self.page3.mirrored = True
        self.page3.rotation_angle = 42
        self.page3.save()

        # The elements retrieved when listing dataset elements are ordered by element id so this
        # test needs to be insensitive to element id changes when regenerating test fixtures
        elements = {
            self.page1.id: {
                "set": "training",
                "element": {
                    "id": str(self.page1.id),
                    "confidence": None,
                    "type": "page",
                    "mirrored": False,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page1.name,
                    "zone": {
                        "id": str(self.page1.id),
                        "image": {
                            "id": str(self.page1.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img1",
                            "path": "img1",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img1/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 0,
                    "classes": None,
                    "has_children": None,
                    "metadata": None,
                    "worker_run": None,
                    "worker_version_id": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.page2.id: {
                "set": "validation",
                "element": {
                    "id": str(self.page2.id),
                    "confidence": None,
                    "type": "page",
                    "mirrored": False,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page2.name,
                    "zone": {
                        "id": str(self.page2.id),
                        "image": {
                            "id": str(self.page2.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img2",
                            "path": "img2",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img2/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 0,
                    "classes": None,
                    "has_children": None,
                    "metadata": None,
                    "worker_run": None,
                    "worker_version_id": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.page3.id: {
                "set": "validation",
                "element": {
                    "id": str(self.page3.id),
                    "confidence": 0.42,
                    "type": "page",
                    "mirrored": True,
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "name": self.page3.name,
                    "zone": {
                        "id": str(self.page3.id),
                        "image": {
                            "id": str(self.page3.image.id),
                            "height": 1000,
                            "width": 1000,
                            "url": "http://server/img3",
                            "path": "img3",
                            "status": "unchecked",
                            "s3_url": None,
                            "server": {
                                "display_name": "Test Server",
                                "max_height": None,
                                "max_width": None,
                                "url": "http://server"
                            },
                        },
                        "polygon": [
                            [0, 0],
                            [0, 1000],
                            [1000, 1000],
                            [1000, 0],
                            [0, 0]
                        ],
                        "url": "http://server/img3/0,0,1000,1000/full/0/default.jpg"
                    },
                    "rotation_angle": 42,
                    "classes": None,
                    "has_children": None,
                    "metadata": None,
                    "worker_run": None,
                    "worker_version_id": None,
                    "thumbnail_url": None,
                    "created": FAKE_CREATED
                },
            },
            self.vol.id: {
                "set": "test",
                "element": {
                    "id": str(self.vol.id),
                    "confidence": None,
                    "type": "volume",
                    "mirrored": False,
                    "name": "Volume 1",
                    "corpus": {
                        "id": str(self.corpus.id),
                        "name": "Unit Tests",
                        "public": True
                    },
                    "zone": None,
                    "rotation_angle": 0,
                    "classes": None,
                    "has_children": None,
                    "metadata": None,
                    "worker_run": None,
                    "worker_version_id": None,
                    "thumbnail_url": "s3_url",
                    "created": FAKE_CREATED
                },
            }
        }

        first_three = [value for key, value in sorted(elements.items())[:3]]
        self.client.force_login(self.read_user)
        for state in DatasetState:
            # A dataset should be readable in any state
            self.dataset.state = state
            self.dataset.save()
            with self.subTest(state=state):
                with self.assertNumQueries(6):
                    response = self.client.get(
                        reverse("api:dataset-elements", kwargs={"pk": self.dataset.pk}),
                        {"page_size": 3},
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                data = response.json()
                self.assertTrue("?cursor=" in data["next"])
                self.assertIsNone(data["count"])
                self.assertListEqual(data["results"], first_three)

    def test_add_element_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_add_element_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_add_element_private_dataset(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_add_element_requires_writable_corpus(self):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            "detail": "You do not have contributor access to the corpus of this dataset."
        })

    def test_add_element_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_id": ["This field is required."],
            "set": ["This field is required."],
        })

    def test_add_element_wrong_element(self):
        element = self.private_corpus.elements.create(type=self.private_corpus.types.create(slug="folder"))
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(element.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "element_id": [f'Invalid pk "{element.id}" - object does not exist.'],
        })

    def test_add_element_wrong_set(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "aaaaaaaaaaa", "element_id": str(self.vol.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set": ["This dataset has no set named aaaaaaaaaaa."],
        })

    def test_add_element_dataset_requires_open(self):
        self.client.force_login(self.user)
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.vol.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertListEqual(response.json(), ["You can only add elements to a dataset in an open state."])

    def test_add_element_already_exists(self):
        self.dataset.dataset_elements.create(element=self.page1, set="test")
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "test", "element_id": str(self.page1.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"non_field_errors": ["This element is already part of this set."]})

    def test_add_element(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(13):
            response = self.client.post(
                reverse("api:dataset-elements", kwargs={"pk": self.dataset.id}),
                data={"set": "training", "element_id": str(self.page1.id)},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertQuerysetEqual(
            self.dataset.dataset_elements.values_list("set", "element__name").order_by("element__name"),
            [("training", "Volume 1, page 1r")]
        )

    def test_add_from_selection_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_add_from_selection_forbidden_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("get", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                response = getattr(self.client, method)(
                    reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id})
                )
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_add_from_selection_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_add_from_selection_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.private_corpus.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_add_from_selection_requires_writable_corpus(self):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            "detail": "You need a Contributor access to the corpus to perform this action."
        })

    def test_add_from_selection_required_fields(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "dataset_id": ["This field is required."],
            "set": ["This field is required."],
        })

    def test_add_from_selection_wrong_values(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set": {}, "dataset_id": "AAA"},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "dataset_id": ["“AAA” is not a valid UUID."],
            "set": ["Not a valid string."],
        })

    def test_add_from_selection_wrong_dataset(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set": "aaa", "dataset_id": self.private_dataset.id},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "dataset_id": [f"Dataset {self.private_dataset.id} is not part of corpus Unit Tests."],
        })

    def test_add_from_selection_completed_dataset(self):
        """A dataset in the Complete state is immutable"""
        self.client.force_login(self.user)
        self.dataset.state = DatasetState.Complete
        self.dataset.save()
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set": "aaa", "dataset_id": self.dataset.id},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "dataset_id": [f"Dataset {self.dataset.id} is marked as completed."]
        })

    def test_add_from_selection_wrong_set(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set": "aaa", "dataset_id": self.dataset.id},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "set": ["This dataset only allows one of training, test, validation."],
        })

    def test_add_from_selection(self):
        self.dataset.dataset_elements.create(element=self.page1, set="training")
        self.assertQuerysetEqual(
            self.dataset.dataset_elements.values_list("set", "element__name").order_by("element__name"),
            [("training", "Volume 1, page 1r")]
        )
        self.user.selected_elements.set([self.vol, self.page1, self.page2])

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:dataset-elements-selection", kwargs={"pk": self.corpus.id}),
                data={"set": "training", "dataset_id": self.dataset.id},
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertQuerysetEqual(
            self.dataset.dataset_elements.values_list("set", "element__name").order_by("element__name"),
            [
                ("training", "Volume 1"),
                ("training", "Volume 1, page 1r"),
                ("training", "Volume 1, page 1v"),
            ]
        )

    def test_element_datasets_requires_read_access(self):
        self.client.force_login(self.user)
        private_elt = self.private_corpus.elements.create(type=self.private_corpus.types.create(slug="t"), name="elt")
        with self.assertNumQueries(5):
            response = self.client.get(reverse("api:element-datasets", kwargs={"pk": private_elt.id}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_element_datasets_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("post", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                client_method = getattr(self.client, method)
                response = client_method(reverse("api:element-datasets", kwargs={"pk": str(self.vol.id)}))
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_element_datasets_public(self):
        """
        A non authenticated user can list datasets of a public element
        """
        self.dataset.dataset_elements.create(element=self.vol, set="train")
        with self.assertNumQueries(3):
            response = self.client.get(reverse("api:element-datasets", kwargs={"pk": str(self.vol.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 1,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
            }]
        })

    def test_element_datasets(self):
        self.client.force_login(self.user)
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.dataset.dataset_elements.create(element=self.page1, set="validation")
        self.dataset2.dataset_elements.create(element=self.page1, set="train")
        with self.assertNumQueries(7):
            response = self.client.get(reverse("api:element-datasets", kwargs={"pk": str(self.page1.id)}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            "count": 3,
            "next": None,
            "number": 1,
            "previous": None,
            "results": [{
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
            }, {
                "dataset": {
                    "id": str(self.dataset.id),
                    "name": "First Dataset",
                    "description": "dataset number one",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "created": self.dataset.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "validation",
            }, {
                "dataset": {
                    "id": str(self.dataset2.id),
                    "name": "Second Dataset",
                    "description": "dataset number two",
                    "sets": ["training", "test", "validation"],
                    "set_elements": None,
                    "state": "open",
                    "corpus_id": str(self.corpus.id),
                    "creator": "Test user",
                    "task_id": None,
                    "created": self.dataset2.created.isoformat().replace("+00:00", "Z"),
                    "updated": self.dataset2.updated.isoformat().replace("+00:00", "Z"),
                },
                "set": "train",
            }]
        })

    # DatasetClone

    def test_clone_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "Authentication credentials were not provided."})

    def test_clone_forbidden_methods(self):
        self.client.force_login(self.user)
        forbidden_methods = ("get", "patch", "put", "delete")
        for method in forbidden_methods:
            with self.subTest(method=method):
                response = getattr(self.client, method)(
                    reverse("api:dataset-clone", kwargs={"pk": self.dataset.id})
                )
                self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_clone_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You do not have permission to perform this action."})

    def test_clone_private_corpus(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.private_dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_clone_requires_writable_corpus(self):
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            "detail": "You need a Contributor access to the dataset's corpus."
        })

    def test_clone(self):
        self.dataset.creator = self.superuser
        self.dataset.state = DatasetState.Error
        self.dataset.task = self.task
        self.dataset.save()
        self.dataset.dataset_elements.create(element=self.page1, set="test")
        self.dataset.dataset_elements.create(element=self.page1, set="validation")
        self.dataset.dataset_elements.create(element=self.vol, set="validation")
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), ["First Dataset", "Second Dataset"])

        self.client.force_login(self.user)
        with self.assertNumQueries(15):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}),
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), [
            "First Dataset",
            "Clone of First Dataset",
            "Second Dataset",
        ])
        data = response.json()
        clone = self.corpus.datasets.get(id=data["id"])
        self.assertEqual(clone.creator, self.user)
        data.pop("created")
        data.pop("updated")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(clone.id),
                "name": "Clone of First Dataset",
                "description": self.dataset.description,
                "creator": self.user.display_name,
                "corpus_id": str(self.corpus.id),
                "sets": ["training", "test", "validation"],
                "set_elements": {"test": 1, "training": 0, "validation": 2},
                "state": DatasetState.Open.value,
                "task_id": str(self.task.id),
            },
        )
        self.assertQuerysetEqual(
            clone.dataset_elements.values_list("set", "element__name").order_by("element__name", "set"),
            [
                ("validation", "Volume 1"),
                ("test", "Volume 1, page 1r"),
                ("validation", "Volume 1, page 1r"),
            ]
        )

    def test_clone_existing_name(self):
        self.corpus.datasets.create(name="Clone of First Dataset", creator=self.user)
        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": self.dataset.id}),
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(self.corpus.datasets.values_list("name", flat=True), [
            "First Dataset",
            "Clone of First Dataset",
            "Clone of First Dataset 1",
            "Second Dataset",
        ])

        data = response.json()
        data.pop("id")
        data.pop("created")
        data.pop("updated")
        self.assertDictEqual(
            response.json(),
            {
                "name": "Clone of First Dataset 1",
                "description": self.dataset.description,
                "creator": self.user.display_name,
                "corpus_id": str(self.corpus.id),
                "sets": self.dataset.sets,
                "set_elements": {k: 0 for k in self.dataset.sets},
                "state": DatasetState.Open.value,
                "task_id": None,
            },
        )

    def test_clone_name_too_long(self):
        dataset = self.corpus.datasets.create(name="A" * 99, creator=self.user)
        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse("api:dataset-clone", kwargs={"pk": dataset.id}),
                format="json",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        clone = Dataset.objects.get(id=data.pop("id"))
        self.assertEqual(clone.name, "Clone of " + "A" * 91)
        self.assertEqual(data["name"], clone.name)

    # DatasetElementDestroy

    def test_destroy_dataset_element_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_dataset_element_requires_verified(self):
        user = User.objects.create(email="not_verified@mail.com", display_name="Not Verified", verified_email=False)
        self.client.force_login(user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_destroy_dataset_element_requires_contributor(self):
        self.client.force_login(self.read_user)
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.dataset.dataset_elements.create(element=self.page1, set="validation")
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 1)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {"detail": "You need a Contributor access to the dataset to perform this action."})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 1)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)

    def test_destroy_dataset_element_set_required(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"set": ["This field is required."]})

    def test_destroy_dataset_element_requires_open_dataset(self):
        self.client.force_login(self.user)
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.dataset.dataset_elements.create(element=self.page1, set="validation")
        self.dataset.state = DatasetState.Error
        self.dataset.save()
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 1)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"dataset": ["Elements can only be removed from open Datasets."]})
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 1)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)

    def test_destroy_dataset_element_dataset_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_destroy_dataset_element_set_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=match"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_destroy_dataset_element_element_doesnt_exist(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_destroy_dataset_element_element_not_in_dataset(self):
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page2.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_destroy_dataset_element_wrong_set(self):
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.dataset.dataset_elements.create(element=self.page2, set="validation")
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page2.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {"detail": "Not found."})

    def test_destroy_dataset_element(self):
        self.client.force_login(self.user)
        self.dataset.dataset_elements.create(element=self.page1, set="train")
        self.dataset.dataset_elements.create(element=self.page1, set="validation")
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 1)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse(
                "api:dataset-element",
                kwargs={"dataset": str(self.dataset.id), "element": str(self.page1.id)})
                + "?set=train"
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.dataset.refresh_from_db()
        self.assertEqual(self.dataset.dataset_elements.filter(set="train").count(), 0)
        self.assertEqual(self.dataset.dataset_elements.filter(set="validation").count(), 1)
